import makeEventHandler from "./events.js";

function makeRenderer(canvas = {}) {
  const visibleContext = canvas.getContext("2d");
  if (!visibleContext) {
    throw "Impossible to make renderer. Can not get canvas 2d context.";
  }

  const events = makeEventHandler();

  const layers = new Map();

  resize();

  function resize() {
    // Using parseInt because window.devicePixelRatio is a float and can return half pixels
    const viewportWidth = parseInt(
      canvas.clientWidth * window.devicePixelRatio
    );
    const viewportHeight = parseInt(
      canvas.clientHeight * window.devicePixelRatio
    );
    if (canvas.width !== viewportWidth || canvas.height !== viewportHeight) {
      const dx = viewportWidth - canvas.width;
      const dy = viewportHeight - canvas.height;
      resizeAll(viewportWidth, viewportHeight);
      events.emit("resize", { dx, dy });
    }
  }

  function resizeAll(width, height) {
    getContexts().forEach(context => {
      context.canvas.width = width;
      context.canvas.height = height;
    });
  }

  function getContexts() {
    return [visibleContext, ...getLayers().map(layer => layer.context)];
  }

  function getLayers() {
    return Array.from(layers.values());
  }

  function makeLayer(objects, visible, batchPath) {
    return {
      _objects: objects,
      get objects() {
        return this._objects;
      },
      set objects(newObjects) {
        this.rendered = false;
        if (this.batchPath) {
          this.materialsBatchPath.clear();
        }
        this._objects = newObjects;
      },
      visible,
      context: setOffscreenContext(canvas.width, canvas.height),
      rendered: false,
      setTransform(transform) {
        this.rendered = false;
        this.context.setTransform(transform);
      },
      batchPath,
      materialsBatchPath: batchPath ? new Map() : undefined,
    };
  }

  function drawLayer(layer) {
    visibleContext.drawImage(layer.context.canvas, 0, 0);
    return layer;
  }

  return {
    transform: visibleContext.getTransform(),
    transformChanged: true,
    setTransform(t) {
      if (this.transform !== t) {
        this.transform = t;
        this.transformChanged = true;
      }
    },
    getViewport() {
      return {
        get width() {
          return canvas.width;
        },
        get height() {
          return canvas.height;
        },
      };
    },
    addLayer(name, objects, visible = true, batchPath = true) {
      const layer = makeLayer(objects, visible, batchPath);
      layers.set(name, layer);
      return layer;
    },
    setLayerObjects(name, objects) {
      if (!layers.has(name)) {
        throw `Layer "${name}" does not exist.`;
      }
      const layer = layers.get(name);
      layer.objects = objects;
    },
    render() {
      resize();
      if (this.transformChanged) {
        getLayers().forEach(layer => layer.setTransform(this.transform));
        this.transformChanged = false;
      }

      const layersToRedraw = getLayers().filter(layer => !layer.rendered);

      if (layersToRedraw.length > 0) {
        layersToRedraw.map(clearLayer).forEach(renderLayer);

        clear(visibleContext);

        getLayers()
          .filter(layer => layer.visible && layer.objects.length)
          .forEach(drawLayer);
      }
    },
    ...events,
  };
}

function setOffscreenContext(width, height) {
  const offscreenCanvas = document.createElement("canvas");
  offscreenCanvas.width = width;
  offscreenCanvas.height = height;
  return offscreenCanvas.getContext("2d");
}

function makeBatchPath(objects) {
  const batchPath = new Path2D();
  objects.forEach(object => batchPath.addPath(object.path));
  return batchPath;
}

function clearLayer(layer) {
  clear(layer.context);
  return layer;
}

function clear(context) {
  context.save();
  context.resetTransform();
  context.clearRect(0, 0, context.canvas.width, context.canvas.height);
  context.restore();
  return context;
}

function renderLayer(layer) {
  if (layer.batchPath) {
    const objectsByMaterial = new Map();
    layer.objects.forEach(object => {
      if (objectsByMaterial.has(object.material)) {
        objectsByMaterial.get(object.material).push(object);
      } else {
        objectsByMaterial.set(object.material, [object]);
      }
    });
    objectsByMaterial.forEach((objects, material) => {
      let batchPath = null;
      if (layer.materialsBatchPath.has(material)) {
        batchPath = layer.materialsBatchPath.get(material);
      } else {
        batchPath = makeBatchPath(objects);
        layer.materialsBatchPath.set(material, batchPath);
      }
      renderPath(layer.context, batchPath, material);
    });
  } else {
    layer.objects.forEach(object =>
      renderPath(layer.context, object.path, object.material)
    );
  }

  layer.rendered = true;
  return layer;
}

function renderPath(context, path, material) {
  setContextMaterial(context, material);
  if (material.strokeStyle) {
    context.stroke(path);
  }
  if (material.fillStyle) {
    context.fill(path);
  }
}

function setContextMaterial(context, material) {
  context.lineWidth = material.strokeStyle && material.strokeStyle.lineWidth;
  context.strokeStyle = material.strokeStyle && material.strokeStyle.color;
  context.fillStyle = material.fillStyle && material.fillStyle.color;
}

export default makeRenderer;
