import { materials as defaultMaterials } from "./materials.js";
import {
  getRGBStringColorFromId,
  getIdFromRGBColor,
  fromUint8ArrayColorToStringColor,
} from "./utils.js";

function makeStateRenderAdapter(state, renderer) {
  const materials = { ...defaultMaterials };

  const parsedMaterials = new Map();

  function parseMaterial(material = {}) {
    if (parsedMaterials.has(material)) {
      return parsedMaterials.get(material);
    } else {
      const parsedMaterial = {};
      if (material.strokeStyle) {
        parsedMaterial.strokeStyle = {
          lineWidth: material.strokeStyle.lineWidth,
          color: fromUint8ArrayColorToStringColor(
            material.strokeStyle.color,
            material.strokeStyle.opacity
          ),
        };
      }
      if (material.fillStyle) {
        parsedMaterial.fillStyle = {
          color: fromUint8ArrayColorToStringColor(
            material.fillStyle.color,
            material.fillStyle.opacity
          ),
        };
      }
      parsedMaterials.set(material, parsedMaterial);
      return parsedMaterial;
    }
  }

  function makeAdapterLayer(cfg) {
    const { name, objects = [], visible, batchPath, materialCallback } = cfg;
    const rendererLayer = renderer.addLayer(name, objects, visible, batchPath);

    return new Proxy(
      {
        rendererLayer,
        objects: {},
        removeObjects(objects) {
          let someObjectDeleted = false;
          objects.forEach(object => {
            if (this.objects[object.id] !== undefined) {
              delete this.objects[object.id];
              someObjectDeleted = true;
            }
          });
          if (someObjectDeleted) {
            this.updateRenderer();
          }
        },
        addObjects(objects) {
          objects.forEach(object => {
            this.objects[object.id] = {
              rank: -object.area,
              path: object.path,
              material: materialCallback(object),
            };
          });
          this.updateRenderer();
        },
        updateRenderer() {
          renderer.setLayerObjects(
            name,
            Object.values(this.objects).sort((a, b) =>
              a.rank > b.rank ? 1 : a.rank < b.rank ? -1 : 0
            )
          );
        },
      },
      {
        // automatic bind methods
        get(target, key) {
          return typeof target[key] === "function"
            ? target[key].bind(target)
            : target[key];
        },
      }
    );
  }

  function getMaterialByObjectType(type) {
    return Object.values(materials)
      .filter(material => material.types)
      .find(material => material.types.includes(type));
  }

  const mainAdapterLayer = makeAdapterLayer({
    name: "main",
    materialCallback: object =>
      parseMaterial(getMaterialByObjectType(object.type) || materials.default),
  });
  const pickingAdapterLayer = makeAdapterLayer({
    name: "picking",
    visible: false,
    batchPath: false,
    materialCallback: object => ({
      fillStyle: { color: getRGBStringColorFromId(object.id) },
    }),
  });
  const selectedObjectsAdapterLayer = makeAdapterLayer({
    name: "select",
    materialCallback: () => parseMaterial(materials.selected),
  });
  const highlightedObjectsAdapterLayer = makeAdapterLayer({
    name: "highlight",
    materialCallback: () => parseMaterial(materials.highlighted),
  });

  state.on("objects-added", onObjectsAdded);

  state.on("objects-removed", onObjectsRemoved);

  state.on("objects-selected", objects =>
    selectedObjectsAdapterLayer.addObjects(objects.filter(o => o.visible))
  );

  state.on("objects-unselected", selectedObjectsAdapterLayer.removeObjects);

  state.on("objects-highlighted", objects =>
    highlightedObjectsAdapterLayer.addObjects(objects.filter(o => o.visible))
  );

  state.on(
    "objects-unhighlighted",
    highlightedObjectsAdapterLayer.removeObjects
  );

  state.on("objects-pickable", objects =>
    pickingAdapterLayer.addObjects(objects.filter(o => o.visible))
  );

  state.on("objects-unpickable", pickingAdapterLayer.removeObjects);

  state.on("objects-visible", objects => {
    mainAdapterLayer.addObjects(objects);
    pickingAdapterLayer.addObjects(objects.filter(object => object.pickable));
    selectedObjectsAdapterLayer.addObjects(
      objects.filter(object => object.selected)
    );
    highlightedObjectsAdapterLayer.addObjects(
      objects.filter(object => object.highlighted)
    );
  });

  state.on("objects-unvisible", objects => {
    removeObjectsOfAllAdapterLayers(objects);
  });

  function onObjectsAdded(objects) {
    const visibleObjects = objects.filter(object => object.visible);
    mainAdapterLayer.addObjects(visibleObjects);
    pickingAdapterLayer.addObjects(
      visibleObjects.filter(object => object.pickable)
    );
    selectedObjectsAdapterLayer.addObjects(
      visibleObjects.filter(object => object.selected)
    );
    highlightedObjectsAdapterLayer.addObjects(
      visibleObjects.filter(object => object.highlighted)
    );
  }

  function onObjectsRemoved(objects) {
    removeObjectsOfAllAdapterLayers(objects);
  }

  function removeObjectsOfAllAdapterLayers(objects) {
    [
      mainAdapterLayer,
      pickingAdapterLayer,
      selectedObjectsAdapterLayer,
      highlightedObjectsAdapterLayer,
    ].forEach(adapterLayer => adapterLayer.removeObjects(objects));
  }

  return {
    materials,
    getObjectAtPosition(position) {
      const pickingContext = pickingAdapterLayer.rendererLayer.context;
      const { data: color } = pickingContext.getImageData(
        position.x,
        position.y,
        1,
        1
      );
      const id = getIdFromRGBColor(color);
      const object = state.getObject(id);
      return object;
    },
  };
}

export default makeStateRenderAdapter;
